/*
	Car Gauges v0.2
	Chris Hii

	First version:	Nov 9th, 2010
	Last revised: 	Apr 21st, 2011

	PORTA.0 - Oil Pressure Sender
	PORTA.1 - Unused
	PORTA.2 - 3k/1k Voltage Divider for Battery Voltage
	PORTA.3 - Unused

	Use LM35 on stock temp gauge port? 12mm x 1.25

	DS18B20's
		0: 28 93 E5 DF 02 00 00 2A
		1: 28 F8 CF DF 02 00 00 54
		2: 28 F8 CF DF 02 00 00 54
		3: 28 9D 46 E1 02 00 00 19
		4: 28 47 5A E1 02 00 00 4F
		5: 28 2C 5D E1 02 00 00 25
		6: 28 CA 6B E1 02 00 00 A6
		7: 28 97 7B E1 02 00 00 BE
*/

#pragma chip PIC16F877A
//#pragma config = 0x2F8A
#pragma config = 0x2F92	// brown out reset and power up timer enabled

#pragma bit PWR_LED @ PORTD.0
#pragma bit STATUS_LED @ PORTD.1
#pragma bit LOW_PRES_LED @ PORTD.2

#define NO_DECIMAL 0
#define DECIMAL 1
#define CLT_TEMP_SENSOR 0
#define INTAKE_TEMP_SENSOR 1
#define PRESSURE_CHANNEL 0										// oil pressure on an0
#define VOLTAGE_CHANNEL 2										// voltage divider on an2
#define CLOCK 20000000
#define LOW_PRES_THESH 12										// low oil psi threshold
#define VOLTAGE_MULTIPLIER 200.0								// float dependent on regulator output and voltage divider resistors

// 0x80		[ Oil Pres   42  psi ]
// 0xC0		[  Coolant   91.5  C ]
// 0x94		[   Intake   30.5  C ]
// 0xD4		[  Battery   14.3  V ]

// 0x80		[ Oil  42    psi     ]
// 0xC0		[ Clt  91.5  C       ]
// 0x94		[ IAT  30.5  C       ]
// 0xD4		[ Bat  14.3  V       ]

#define LINE0 0x80
#define LINE1 0xC0
#define LINE2 0x94
#define LINE3 0xD4

// Positions of display values
#define OIL_PRES_LINE	0
#define OIL_PRES_POS	11
#define CLT_TEMP_LINE	1
#define CLT_TEMP_POS	11
#define IAT_TEMP_LINE	2
#define IAT_TEMP_POS	11
#define VOLTAGE_LINE	3
#define VOLTAGE_POS	12


static const uns8 static_text[] = " Oil Pres       psi "
								  "  Coolant         C "
								  "   Intake         C "
								  "  Battery         V ";

/*
	Defines for pressure sender. Using ProSport Premium oil pressure sender (0~5V, 0-145psi).
*/
#define MIN_PRES_ADC 0
#define MAX_PRES_ADC 1023
#define RANGE_PRES_ADC 1023
#define MIN_PRES 0
#define MAX_PRES 145
#define ATMOSPHERIC_PRESS 14.7
/*
	16-bit float functions, delays, and OneWire library.
*/
#include <MATH16F.h>
#include "20mhz_delays.h"
#include "ds18b20.h"

/*
    Really stupid, but have to include the .c files as well as the .h files.
*/
#include "20mhz_delays.c"
#include "ds18b20.c"

uns8 digitOne, digitTwo, digitThree,
	 state;

void init(void);
uns8 calc_pres(void);
uns8 calc_voltage(void);
uns16 get_adc_value(uns8);
void move_cursor(uns8, uns8);
void print_static_text(void);
void hex_to_ascii(uns8);
void send_byte(uns8);
void send_str(const uns8 *);
void send_cmd(uns8);
void send_value(uns8, bit);
void low_pressure(void);
uns8 process_temp(uns8 the_temp);			// handle temperature displays, including negative signs and decimals

void main(void)
{
	uns8 oil_pressure,
		 coolant_temp,
		 intake_temp,
		 voltage;

	init();

	while(1) {

		STATUS_LED = 1;

		switch(state) {
			case 0:
				onewire_start_conv(0);							// perform conversion once every 8 main loops (~4 seconds)
				state++;
				break;
			case 7:
				//print_static_text();							// redraw static text every 4 seconds to clean up screen
				state = 0;										// restart state counter
				break;
			default:
				state++;
				break;
		}

		oil_pressure = calc_pres();
		voltage = calc_voltage();

		//if (oil_pressure < LOW_PRES_THESH) {					// check for low pressure
		//	low_pressure();
		//	continue;
		//}
		//else
		//	LOW_PRES_LED = 0;

		move_cursor(OIL_PRES_LINE, OIL_PRES_POS);
		send_value(oil_pressure, NO_DECIMAL);									// display oil pressure

		move_cursor(CLT_TEMP_LINE, CLT_TEMP_POS);
		coolant_temp = process_temp(ds18b20_read_temp(CLT_TEMP_SENSOR));		// display coolant temp

		move_cursor(IAT_TEMP_LINE, IAT_TEMP_POS);
		intake_temp = process_temp(ds18b20_read_temp(INTAKE_TEMP_SENSOR));		// display intake temp

		move_cursor(VOLTAGE_LINE, VOLTAGE_POS);
		send_value(voltage, DECIMAL);											// display supply voltage

	// 0x80		[ Oil Pres   42  psi ]
	// 0xC0		[  Coolant   91.5  C ]
	// 0x94		[   Intake   30    C ]
	// 0xD4		[  Battery   14.3  V ]

		delay_ms(150);
		STATUS_LED = 0;											// use green led as heartbeat (running indicator)
		delay_ms(250);											// screen refresh rate
	}
}

void init(void)
{
	TRISA = 0x1F;
	TRISD = 0x00;
	SPBRG = 31; 												// 9600 baud at 20MHz
	SYNC = 0;
	SPEN = 1;
	TXSTA = 0xA0;
	ADCON1 = 0x82; 												// right justified, AN0 AN1 AN2 AN3 AN4 analog inputs
	ADCON0 = 0x81; 												// Fosc/32 conversion, AN0, A/D enabled

	delay_ms(3000);

	print_static_text();										// fill display with static text

	state = 0;													// start in initial state

    PWR_LED = 1;
	STATUS_LED = 1;												// turn on green led
}

/*
	Calculates pressure in psig using voltage transducer.
*/
uns8 calc_pres(void)
{
	float16 adc_float;

	adc_float = (float16)get_adc_value(PRESSURE_CHANNEL);

	if (adc_float < MIN_PRES_ADC)			 					// pressure below scale
		return MIN_PRES;										// zero

	adc_float -= MIN_PRES_ADC;
	adc_float /= RANGE_PRES_ADC;
	adc_float *= MAX_PRES;

	adc_float -= ATMOSPHERIC_PRESS;								// offset for atmospheric pressure

	if(adc_float < 0)											// cap at 0 psi minimum
		adc_float = 0;

	return (uns8)adc_float;
}

uns8 calc_voltage(void)
{
// returns voltage in three digits, last digit is first decimal value

	float16 adc_float;

	adc_float = (float16)get_adc_value(VOLTAGE_CHANNEL);
	adc_float /= 1023.0;										// find percentage of scale
	adc_float *= VOLTAGE_MULTIPLIER;							// depends on voltage reference

	return (uns8)adc_float;										// cast to unsigned and return
}

uns16 get_adc_value(uns8 channel)
{
	uns16 adc_val;

	ADCON0 &= 0xC5;												// clear a/d channel select
	channel = channel << 3;
	ADCON0 |= channel;											// select channel
	ADCON0.2 = 1;												// begin a/d conversion
	while(ADCON0.2);											// wait for conversion to finish

	adc_val = ADRESH * 256;
	adc_val += ADRESL;

	return adc_val;
}

void move_cursor(uns8 line_num, uns8 pos)
{
	uns8 byte;

	switch(line_num) {
		case 0:
			byte = LINE0;
			break;
		case 1:
			byte = LINE1;
			break;
		case 2:
			byte = LINE2;
			break;
		case 3:
			byte = LINE3;
			break;
	}

	byte += pos;
	send_cmd(byte);
}

void print_static_text(void)
{
	move_cursor(0, 0);
	send_str(static_text);
}

void hex_to_ascii(uns8 number)
{
	uns8 var;
	digitOne = digitTwo = digitThree = 0;

	digitOne = number / 100;
	var = digitOne * 100;
	digitTwo = number - var;
	digitTwo /= 10;
	digitThree = number - var;
	var = digitTwo * 10;
	digitThree -= var;

	digitOne += 0x30;											// offsets for ascii numbers
	digitTwo += 0x30;
	digitThree += 0x30;

}

void send_byte(uns8 byte_to_send)
{
	TXREG = byte_to_send;
	while(!TRMT);
}

void send_str(const uns8 * string)
{
	uns8 ptr = 0;
	while(string[ptr] != '\0')
		send_byte(string[ptr++]);
}

void send_cmd(uns8 cmd)
{
	send_byte(0xFE);											// command byte
	send_byte(cmd);
}

void send_value(uns8 value, bit decimal)
{
	hex_to_ascii(value);

	if (digitOne == 0x30)										// strip leading zeros
		send_str(" ");
	else
		send_byte(digitOne);

	if (digitOne == 0x30 && digitTwo == 0x30)					// strip leading zeros
		send_str(" ");
	else
		send_byte(digitTwo);

	if (decimal)
		send_str(".");

	send_byte(digitThree);

}

void low_pressure(void)
{
	LOW_PRES_LED = 1;
	send_cmd(0x01);												// clear lcd and flash warning
	send_str("  LOW OIL PRESSURE  ");
	delay_ms(200);
	send_cmd(0x01);
	delay_ms(50);
}

uns8 process_temp(uns8 the_temp)
{
	if (the_temp & 0x80) {
		send_str("-");
		the_temp &= 0x7F;
	}

	send_value(the_temp, NO_DECIMAL);							// send whole number of temperature

	if (half_degree)											// if half degree is present, send as well
		send_str(".5");
	else
		send_str("  ");

	return the_temp;
}
